/*
 * Filename: PQHILITE.C Vs 1.4
 * Author:   Andras Hoeffken <ah@confusion.rmc.de> - 2:2480/13.34 @ fidonet
 * Status:   released to public domain
 * History:  September 20, 1996
 *
 * Runs once through the video text RAM, picks out each word, compares it
 * with the list of keywords and if found, highlights it with individual
 * color.
 * In addition: looks for specific lines like 'REMarks', 'inline ASM' etc.
 *
 * compile command: BCC -ms -O2 -Og -w -w-param -d -K -c FILENAME.C
 *   (Borland C++
 *     Vs 3.1)      the switch O2 is VERY important: high speed performance
 *                    (40 % faster than O1 !!)
 *
 * Note: as we work with an own ASM root module, the standard C-libraries
 *         can NOT be linked!
 */
/*-----------------------------------------------------*/
#define TRUE  1
#define FALSE 0

#define BSL   92

#define STRING_ATTR   1
#define CONSTANT_ATTR 2
#define POINTER_ATTR  3
#define COMMENT_ATTR  4
#define ASM_ATTR      5
#define KEYWORD_ATTR  6
#define COMPILER_ATTR 7
#define INTERN_ATTR   8
#define USER_ATTR     9

/* from PQSYNTAX.ASM: */
extern unsigned char far *g_phys_screen;   /* an external pointer variable */
extern unsigned char g_k_colors[],g_temp_str[];
extern unsigned char g_newattr;
extern int g_temp_cell, g_current_cell,g_num_lines;

/* from PQSYNTAX.ASM: */
void check_lines(void);
void hilite_string(void);
void remark_string(void);
int array_scan(void);

static int temp_len=0, key_wd_count=0;
static int max_cell=3998;                     /* do tests with 25*80 screen */
static int pb_ide=FALSE,qb_ide=FALSE;         /* assume this for start */


static void markattr(void)
{
    g_temp_cell = g_current_cell - temp_len*2;
    hilite_string();                          /* into backward direction */
    temp_len = 0;
}

static void remarkline(void)
{
    g_newattr = g_k_colors[COMMENT_ATTR];
    g_temp_cell = (int)(g_current_cell / 160); /* line number */
    g_temp_cell = g_temp_cell*160 + 156;       /* 1 before end of line */
    remark_string();                           /* into forward direction */
    g_current_cell=g_temp_cell+2;
}

static void remarkqbline(void) /* QB: look for embedded metacommands $xxx */
{
    int start_cell,group;
    unsigned char c;

    start_cell = g_current_cell;
    g_newattr = g_k_colors[COMMENT_ATTR];
    markattr();                           /* first hilite the ' or REM */
    g_current_cell=g_current_cell+2;

    g_temp_cell = (int)(g_current_cell / 160); /* line number */
    g_temp_cell = g_temp_cell*160 + 142;       /* 8 before end of line */

    while (g_current_cell <= g_temp_cell){
        c = g_phys_screen[g_current_cell];
        switch (c){
        case ' ':
            break;                  /* run over spaces */
        case '$':
            temp_len=0;
            do {
                if ((c>96)&&(c<123)) c=c-32; /* make upper case */
                temp_len = temp_len + 1;
                g_temp_str[temp_len]=c;      /* collect temp data */
                g_current_cell=g_current_cell+2;
                if (g_current_cell == g_temp_cell+16){
                    g_current_cell=start_cell; remarkline(); return;
                }
                else {

                    c = g_phys_screen[g_current_cell];
                }
            } while (c!=' ');
            g_temp_str[0]=temp_len;       /* the length of the actual word */
            group = array_scan();         /* look if this word is a keyword */
            if (group == COMPILER_ATTR){
                g_newattr = g_k_colors[COMPILER_ATTR];
                markattr();
            }
            else {
                g_current_cell = start_cell;
            }
            remarkline(); return;
        default:
            g_current_cell=start_cell; remarkline(); return;
        }
        g_current_cell=g_current_cell+2;
    }
}

/* we have found a separate word (set apart by delimiters). Here we check,
 * if this word is in the list of keywords, and if so the word is marked
 * with its individual color */

static void change_attr(void)
{
    unsigned char fc;
    int group;

    fc = g_temp_str[1];  /* first check for specific preceding chars */
    if (fc  == '%'|| fc == '&' || (fc>47 && fc<58)){ /* numeric */
        g_newattr = g_k_colors[CONSTANT_ATTR];
        markattr();
        return;
    }
    else if (fc == '@'){
        g_newattr = g_k_colors[POINTER_ATTR];
        markattr();
        return;
    }

    key_wd_count=key_wd_count+1;  /* a flag used for the ASM lines */
    g_temp_str[0]=temp_len;       /* the length of the actual word */
    group = array_scan();         /* look if this word is a keyword */

    switch (group){
    case 0:                       /* group 0: NOT a keyword ! */
        temp_len=0;
        return;
    case 4:
        if (pb_ide == TRUE){
          g_current_cell = g_current_cell - temp_len*2;
          remarkline();
        }
        else {
          remarkqbline();
        }
        temp_len=0;
        return;
    case 7:                       /* QB: only embedded in REMarks */
        if (qb_ide == TRUE){
          temp_len=0;
          return;
        }
        g_newattr = g_k_colors[COMPILER_ATTR];
        break;
    case 6:
        g_newattr = g_k_colors[KEYWORD_ATTR];
        break;

    case 5:
        if (key_wd_count == 1){   /* if found as first keyword on a line */
            g_newattr = g_k_colors[ASM_ATTR];
            markattr();
            while ((g_current_cell%160) > 0){ /* over the rest of the line */
                g_current_cell = g_current_cell+2;
                if (g_phys_screen[g_current_cell] == ';'){ /* asm comment */
                    remarkline();
                }
            }
            g_current_cell = g_current_cell-2;
        }
        else {
            temp_len=0;
        }
        return;

    case 8:
        g_newattr = g_k_colors[INTERN_ATTR];
        break;
    case 9:
        g_newattr = g_k_colors[USER_ATTR];
        break;
    }
    markattr();
}

void hilite_screen(void)
{
    int found_quote=FALSE, str_length=0;
    unsigned char c;

    found_quote=FALSE;           /* reset some flags */
    temp_len=0;
    key_wd_count=0;

    while (g_current_cell <= max_cell){ /* scan the whole editor screen RAM */
        if (found_quote == TRUE){       /* process double quoted (34) strings */
                                        /* if 2nd quote or end of line found: */
            if ((g_phys_screen[g_current_cell]==34)||(g_current_cell%160)==156){
                g_current_cell=g_current_cell+2;
                g_temp_cell=g_current_cell - str_length;
                g_newattr = g_k_colors[STRING_ATTR];
                hilite_string();        /* into backward direction */
                found_quote=FALSE;
            }
            else {                      /* else step over the string */
                str_length=str_length+2;
                g_current_cell=g_current_cell+2;
            }
        }
        else{                           /* process all the rest */
          c = g_phys_screen[g_current_cell];
          if (c == ' '){                /* most probable char first */
            if (temp_len != 0) change_attr(); /* of the previous word */
          }                             /* 2nd probable are alphanumerics */
          else if ((c>64 && c<91) || (c>96 && c<123)) {
            if ((c>96)&&(c<123)) c=c-32; /* make upper case */
            temp_len = temp_len + 1;
            g_temp_str[temp_len]=c;      /* collect temp data */
          }
          else {
            switch (c){                 /* look for a word delimiter */
            case '=':
            case '(':
            case ')':
            case '+':
            case '-':
            case '*':
            case '/':
            case '<':
            case '>':
            case BSL:                   /* \ = backslash */
            case ',':
            case ';':
            case '^':
                if (temp_len != 0) change_attr(); /* of the previous word */
                g_newattr = g_k_colors[KEYWORD_ATTR];
                markattr();
                break;
            case 39:                    /* ' = single quote */
                if (temp_len != 0) change_attr(); /* previous word */
                temp_len=0;                       /* and actual string */
                if (pb_ide == TRUE) remarkline(); else remarkqbline();
                break;
            case ':':                   /* : = statement separator */
                if (temp_len != 0) change_attr();
                temp_len=0;
                g_newattr = g_k_colors[COMPILER_ATTR];
                markattr();
                break;
            case 34:                    /* " first double quote found */
                if (temp_len != 0) change_attr(); /** of the previous word */
                found_quote=TRUE;
                str_length=4;
                break;
     /*       case 0: /** specially for keywords in online help of QB/PDS IDE
                if (temp_len != 0) change_attr(); /** of the previous word
                break;  */
            default:                    /* we are in the mid of a word */
                temp_len = temp_len + 1;
                g_temp_str[temp_len]=c;      /* collect temp data */
                break;
            }
          }
          g_current_cell=g_current_cell+2;
          if (g_current_cell % 160 == 0){  /* begin of new line */
              found_quote=FALSE;           /* reset some flags */
              temp_len=0;
              key_wd_count=0;
              g_current_cell=g_current_cell+2; /* step to next column */
          }
        }
    }
}

/* performe some checks before deciding to start highlighting
 * highlighting is done only:
 * a) the IDE of PowerBASIC or QuickBasic/PDS is found open (test a)
 * b) if the edit window of the corresponding IDE is active, i.e. e.g.
 *      if no menu is pulled down (test b)
 */

void work_on_screen(void){

    int temp_row,first_row,last_row;
    char last_id;

    if (g_phys_screen[10] == 'F'){        /* File menu on 6th position */
       pb_ide = TRUE;
    }
    else if (g_phys_screen[6]=='F' || g_phys_screen[6]=='D'){
       qb_ide = TRUE;                     /* File menu on 4th position */
    }
    else {
       return;
    }

/* test a is OK at this place, so we check the current number of screen lines
 * (25 or 50, the user may change this as he likes)
 */
    check_lines();        /* gives the actual value of g_num_lines */


/* first get number of the starting row */
    if (pb_ide == TRUE){
        first_row = 2;    /* PB edit window always starts at line 3 */
    }
    else {                /* QB start line may vary depending on eval window */
        for (temp_row = 0;temp_row <= g_num_lines;temp_row=temp_row+1){
            if (g_phys_screen[temp_row*160]==218){   /* 218 =  */
                first_row = temp_row+1;
                break;
            }
        }
    }

/* then get number of the last row. 'last row + 1' in each IDE has a
 * significant ID char on the first position */

    if (qb_ide == TRUE) last_id = 32; else last_id = 192; /* space or  */

    for (temp_row = first_row;temp_row <= g_num_lines;temp_row=temp_row+1){
        if (g_phys_screen[temp_row*160]==last_id){ /* look for last row */
                last_row = temp_row-1;
            break;
        }
    }

/* test b: look at 'lastrow + 1', another significant char there tells,
 * wether the edit window is active or e.g. a menu is pulled down. */

    if (pb_ide == TRUE){
        if (g_phys_screen[(last_row+1)*160+4] != 180) return;   /*  menu */
        if (g_phys_screen[322] == 218) return; /**  on line 3 pos  2: help */
        if (g_phys_screen[838] == 201) return; /**  on line 6 pos 20: intro */
        for (temp_row = 168; temp_row <= 266; temp_row = temp_row + 2){
            if (g_phys_screen[temp_row] == 218) return; /**  on line 2: menu */
        }
    }
    else {
        temp_row = (last_row+1)*160+2;
        if (g_phys_screen[temp_row] == 'F') return; /** menu */
        if (g_phys_screen[(first_row)*160+6] == 17) return; /** help screen */
        if (g_phys_screen[temp_row] == 'M') return; /** intro */
    }

    /* PB and QB edit windows start at column 2 and end at column 79 */
    g_current_cell = first_row*160 + 2;
    max_cell = last_row* 160 + 156;

    hilite_screen();                          /* start the real work */
}
