/*
 * Filename: PQREADFL.C Vs 1.4
 * Author:   Andras Hoeffken <ah@confusion.rmc.de> - 2:2480/13.34 @ fidonet
 * Status:   released to public domain
 * History:  September 20, 1996
 *
 * Evaluates the PQSYNTAX.DEF file contents (big buffer), extracts the impor-
 * tant information and puts it into a smaller buffer. Both buffers start at
 * the same address. - If g_test is TRUE, a screen report about the DEF file
 * information is presented, so the user can check his setup definitions.
 *
 * compile command: BCC -ms -O1 -Og -w -w-param -d -K -c
 *   (Borland C++       -zC_INIT -zPDGROUP -zASTACK   (Code --> INIT segment)
 *     Vs 3.1)          -zR_INIT -zSDGROUP -zTSTACK   (Data --> INIT segment)
 *                      -zD_INIT -zGDGROUP -zABTACK   (BSS  --> INIT segment)
 *                      PQREADFL.C
 *
 * Note: - switch O1 gives most compact code - speed is not important for init
 *       - switches -zxxxx used to put the code into the INIT segment, other-
 *           wise this code could not been thrown away after TSR-installation!
 *       - as we work with an own ASM root module, the standard C-libraries
 *           can NOT be linked! Screen printout is done therefore by inline
 *           assembler calls.
 *       - this code is not optimized, but it works and does NOT go resident
 *--------------------------------------------------------------------------*/
#define source_buf g_kwrd_buf       /* DEF file contens, put into the RAM */
#define destin_buf g_kwrd_buf       /* changed to keyword format */

#define TRUE 1
#define FALSE 0
#define CR 13
#define LF 10

/* from PQSYNTAX.ASM: */
extern unsigned char g_kwrd_buf[]; /* keyword array storage area in RAM */
extern unsigned char g_k_colors[]; /* keyword color array area in RAM */
extern int g_f_len;                /* length of :DEF file */
extern int g_maxwords;             /* total number of keywords */
extern int g_bcount;               /* length of keyword buffer in bytes */
extern char g_hot_scan;            /* trigger key: scan code */
extern char g_hot_ctrl;            /* trigger key: shift/ctrl/alt condition */
extern char g_delay_cnt;
extern char g_test;                /* if true, screen output is done */
extern char g_z0[];                          /* 1st method: in ASM module */

static char z1[]="Group 1 (strings   ):  $"; /* 2nd method: in this module */
static char z2[]="Group 2 (%constants):  $";
static char z3[]="Group 3 (@pointers ):  $";
static char z4[]="Group 4 (REMarks   ):  $";
static char z5[]="Group 5 (inline ASM):  $";
static char z6[]="Group 6 (keywords  ):  $";
static char z71[]="Group 7 ($";              /* $ is the DOS string delimiter! */
static char z72[]="compiler ):  $";          /*   can't be part of the string */
static char z8[]="Group 8 (pbvIntern ):  $";
static char z9[]="Group 9 (user words):  $";
static char za[]="foreground = xx  background = xx\r\n$";
static char zb[]="foreground = xx  background = xx   $";
static char zc[]="No of keywords = xxx\r\n$";
static char zd[]="\r\nTotal number of bytes (max 8000) found = xxxx\
 / total number of keywords = xxx\r\n$";

static int src_cnt = 0, dst_cnt = 0, line_cnt, cur_words=0;
static unsigned char cur_word_type,fcol,bcol;
static unsigned char a_line[100];  /* intermediate buffer for keywords */

/* copies a line (delimited either by CR or by CR/LF) from the big buffer
 * to a_line[]. line_cnt is the length of the line (substring)
 */
static int line_input(void)
{
    int result = TRUE;

    for (line_cnt=0; ; ){

        if (source_buf[src_cnt] == CR){
            ++src_cnt;
            if (source_buf[src_cnt] == LF) ++src_cnt;
            break;
        }
        else {
            a_line[line_cnt] = source_buf[src_cnt];
            ++src_cnt;
            ++line_cnt;
        }
    }
    if (src_cnt > g_f_len) result = FALSE; /* EOF (End Of .DEF File) reached */
    return result;
}

/* integer to ASCII: 'value' is put as 'n' digits ASCII to 'position'
   of 'string' (right oriented) */

static void put_n_char (int n, int value, char string[], int position)
{
    char digit;
    signed int rest,i;

    rest = value;

    for (i=n-1;i>=0;i=i-1){
        if (rest>=10){
          digit=rest%10+48; rest=rest/10;
        }
        else if (rest>=0){
          digit=rest+48;rest=-1;
        }
        else {
          digit=32;
        }
        string[position+i]=digit;
    }
}

/* evaluate color attribute out of a DEF file line */

static void get_color(void)
{
    fcol=a_line[14];            /* foreground color */
    if (fcol == 32) fcol = 0; else fcol = fcol-48;
    fcol = fcol*10 + (a_line[15])-48;
    bcol=a_line[30];            /* background color */
    if (bcol == 32) bcol = 0; else bcol = bcol-48;
    bcol = bcol*10 + (a_line[31])-48;
    g_k_colors[cur_word_type]=16*bcol+fcol; /* color attribut byte */
}

/* run through the DEF file line by line until EOF is reached */

void eval_file(void)
{
    unsigned char c;
    int cnt;

    for ( ; ; ){
cont:
        if (line_input() == FALSE){    /* read next line into a_line[] */
            g_bcount = dst_cnt;        /* FALSE --> end of file reached */

            put_n_char(3,cur_words,zc,17);
            if (g_test==TRUE) asm {mov dx,offset zc; mov ah,9; int 0x21}

            put_n_char(4,g_bcount,zd,43);
            put_n_char(3,g_maxwords,zd,77);
            if (g_test==TRUE) asm {mov dx,offset zd; mov ah,9; int 0x21}

            break;
        }
        if (line_cnt == 0) goto cont;     /* empty line */
        if (a_line[0] == ';') goto cont;  /* comment line */

        if (a_line[0] == '*'){            /* a line with a type delimiter */
            cur_word_type=(a_line[1])-48; /* ASCII to binary */
            switch (cur_word_type){
            case 0:                       /* read fundamentals */
                fcol=a_line[12];            /* hot key: scan code */
                if (fcol == 32) fcol = 0; else fcol = fcol-48;
                g_hot_scan = fcol*10 + (a_line[13])-48;
                put_n_char(2,g_hot_scan,g_z0,22);

                fcol=a_line[29];            /* hot key: shift status */
                if (fcol == 32) fcol = 0; else fcol = fcol-48;
                g_hot_ctrl = fcol*10 + (a_line[30])-48;
                put_n_char(2,g_hot_ctrl,g_z0,39);

                fcol=a_line[45];            /* delay count for timer INTR */
                if (fcol == 32) fcol = 0; else fcol = fcol-48;
                g_delay_cnt = fcol*10 + (a_line[46])-48;
                put_n_char(2,g_delay_cnt,g_z0,55);

                if (g_test==TRUE) asm {mov dx,offset g_z0; mov ah,9; int 0x21}
                break;
            case 1:
                get_color();              /* read color attribut */
                put_n_char(2,fcol,za,13);
                put_n_char(2,bcol,za,30);
                if (g_test==TRUE) asm {mov dx,offset z1; mov ah,9; int 0x21}
                if (g_test==TRUE) asm {mov dx,offset za; mov ah,9; int 0x21}
                break;
            case 2:
                get_color();
                put_n_char(2,fcol,za,13);
                put_n_char(2,bcol,za,30);
                if (g_test==TRUE) asm {mov dx,offset z2; mov ah,9; int 0x21}
                if (g_test==TRUE) asm {mov dx,offset za; mov ah,9; int 0x21}
                break;
            case 3:
                get_color();
                put_n_char(2,fcol,za,13);
                put_n_char(2,bcol,za,30);
                if (g_test==TRUE) asm {mov dx,offset z3; mov ah,9; int 0x21}
                if (g_test==TRUE) asm {mov dx,offset za; mov ah,9; int 0x21}
                break;
            case 4:
                get_color();
                put_n_char(2,fcol,zb,13);
                put_n_char(2,bcol,zb,30);
                if (g_test==TRUE) asm {mov dx,offset z4; mov ah,9; int 0x21}
                if (g_test==TRUE) asm {mov dx,offset zb; mov ah,9; int 0x21}
                cur_words=0;
                break;
            case 5:
                put_n_char(3,cur_words,zc,17);
                if (g_test==TRUE) asm {mov dx,offset zc; mov ah,9; int 0x21}
                get_color();
                put_n_char(2,fcol,zb,13);
                put_n_char(2,bcol,zb,30);
                if (g_test==TRUE) asm {mov dx,offset z5; mov ah,9; int 0x21}
                if (g_test==TRUE) asm {mov dx,offset zb; mov ah,9; int 0x21}
                cur_words=0;
                break;
            case 6:
                put_n_char(3,cur_words,zc,17);
                if (g_test==TRUE) asm {mov dx,offset zc; mov ah,9; int 0x21}
                get_color();
                put_n_char(2,fcol,zb,13);
                put_n_char(2,bcol,zb,30);
                if (g_test==TRUE) asm {mov dx,offset z6; mov ah,9; int 0x21}
                if (g_test==TRUE) asm {mov dx,offset zb; mov ah,9; int 0x21}
                cur_words=0;
                break;
            case 7:
                put_n_char(3,cur_words,zc,17);
                if (g_test==TRUE) asm {mov dx,offset zc; mov ah,9; int 0x21}
                get_color();
                put_n_char(2,fcol,zb,13);
                put_n_char(2,bcol,zb,30);
                if (g_test==TRUE) asm {mov dx,offset z71; mov ah,9; int 0x21}
                if (g_test==TRUE) asm {mov dl,36; mov ah,2; int 0x21}
                if (g_test==TRUE) asm {mov dx,offset z72; mov ah,9; int 0x21}
                if (g_test==TRUE) asm {mov dx,offset zb; mov ah,9; int 0x21}
                cur_words=0;
                break;
            case 8:
                put_n_char(3,cur_words,zc,17);
                if (g_test==TRUE) asm {mov dx,offset zc; mov ah,9; int 0x21}
                get_color();
                put_n_char(2,fcol,zb,13);
                put_n_char(2,bcol,zb,30);
                if (g_test==TRUE) asm {mov dx,offset z8; mov ah,9; int 0x21}
                if (g_test==TRUE) asm {mov dx,offset zb; mov ah,9; int 0x21}
                cur_words=0;
                break;
            case 9:
                put_n_char(3,cur_words,zc,17);
                if (g_test==TRUE) asm {mov dx,offset zc; mov ah,9; int 0x21}
                get_color();
                put_n_char(2,fcol,zb,13);
                put_n_char(2,bcol,zb,30);
                if (g_test==TRUE) asm {mov dx,offset z9; mov ah,9; int 0x21}
                if (g_test==TRUE) asm {mov dx,offset zb; mov ah,9; int 0x21}
                cur_words=0;
                break;
            default:
                break;
            }
        }
        else {                                    /* a line with a keyword */
            ++g_maxwords;
            ++cur_words;
            destin_buf[dst_cnt] = (unsigned char)line_cnt;
            ++dst_cnt;                            /* 1st byte = word length */
            destin_buf[dst_cnt] = cur_word_type;
            ++dst_cnt;                            /* 2nd byte = word color */
            for (cnt=0; cnt<line_cnt; cnt=cnt+1){
                c = a_line[cnt];                  /* rest: the word itself */
                if ((c>96)&&(c<123)) c=c-32;      /* make upper case */
                destin_buf[dst_cnt] = c;
                ++dst_cnt;
            }
        }
    }
}
