; Filename: PQSYNTAX.ASM - Vs. 1.4
; Author:   Andras Hoeffken <ah@confusion.rmc.de> - 2:2480/13.34 @ fidonet
; Status:   released to public domain
; History:  September 20, 1996

; comlile command: TASM /z /w /ml PQSYNTAX.ASM
;                  (Borland Turbo-Assembler Vs 3.1)

; link command:
;       TLINK /n /d /c /e /s PQSYNTAX PQHILITE PQREADFL, PQSYNTAX, PQSYNTAX,
;       (Borland Turbo-Linker Vs 5.1)
;---------------------------------------------------------------------------
;PQSYNTAX.EXE is installed by the commands
;            "PQSYNTAX /r" or "LH PQSYNTAX /r" (Load High)
;It can be unloaded by the command
;            "PQSYNTAX /U"
;for full information, simply type "PQSYNTAX"

;ATTENTION: when loaded into UMB RAM, speed performence is reduced by MS-DOS!
;If invoked by a batch file, PQSYNTAX will report error level 0,
; if the loading/unloading could be performed correctly.

;RULES of PQSYNTAX:
; 1) to be started in text mode only (otherwise: unpredictable results)
; 2) the status "mono (25x80) or color text mode (25x80 or 50x80)" is checked
;    at installation of the TSR and each time, the trigger key is operated.
; 3) if in color text mode, the actual number of screen lines (25/50/ is
;    checked permanently.

;By the way, PQSYNTAX demonstrates
; a) the mechanisms of constructing, installing and deinstalling a TSR program
; b) the mechanisms to link the ASM body code of the TSR together with
;    object modules, written in the language "Borland C / Turbo C":
;    - 1st C-module produces resident code, used during TSR status
;    - 2nd C-module produces INIT code (only used for installation purpose)
; c) how a TSR switches the stack back and forth to not overload other stacks
;---------------------------------------------------------------------------
;for own conditional assembling:
TSR         equ   1        ; 0 = for test purpose (run once, TSR not installed)
                           ; 1 = compiled as TSR program
PULSES      equ   0        ; assemble pulses for measurement of time behaviour
INTR_EN_BIT equ   0200h    ; status bit in the processor flag register

VIDEO_ISR   equ   0        ; 1 = if someone likes to use video INTR service

; to improve readability:
B           equ   byte ptr
W           equ   word ptr
D           equ   dword ptr
DG          equ   DGROUP
OS          equ   offset
NP          equ   near ptr
;---------------------------------------------------------------------------
                .RADIX 10       ;default number base = dec

TRUE            equ     1
FALSE           equ     0
LF              EQU     10      ;(define ASCII control codes)
CR              EQU     13

KEYWORD_MEMORY  equ     8192    ;max 8 KB for keyword buffer

VIDEO           equ     10h
DISPLAY_MODE    equ     0Fh
MONO_BASE       equ     0B000h
COLOR_BASE      equ     0B800h

DOS             equ     21h
TERMINATE       equ     4Ch
TSR_ERROR       equ     1
CMD_ERROR       equ     2
DISPLAY_STRING  equ     09h
ENV             EQU     2CH     ;Loc.(in PSP): points to the ENVIRONMENT
LEN_P_ST        EQU     80H     ; "      "   : length of parameter string
FREEMEM         equ     49h

GETVECT         equ     35h
SETVECT         equ     25h
VTRIGGER        equ     10h     ; Video (Trigger) interrupt number
KBD_HARD        equ     9       ; hardware keyboard interrupt number
KBD_BIOS        equ     16h     ; BIOS keyboard interrupt number
TIMER           equ     8       ; timer interrupt number
MUPLEX          equ     2Fh     ; multiplex interrupt number

MAKE_TSR        equ     31h

INT_ENABLE      equ     20h
INT_CTL         equ     20h
;---------------------------------------------------------------------------
;       External References
;functions (from PQHILITE.C):
extrn   _work_on_screen:near    ; check for PB/QB IDE and do hiliting
extrn   _hilite_screen:near     ; highlight screen in any case
extrn   _eval_file:near         ; evaluate DEF file
;=======================================================================
_TEXT   SEGMENT BYTE PUBLIC 'CODE'
_TEXT   ENDS
_DATA   SEGMENT PARA PUBLIC 'DATA'
_DATA   ENDS
_BSS    SEGMENT WORD PUBLIC 'BSS'
_BSS    ENDS
_BSSEND SEGMENT WORD PUBLIC 'STACK'
_BSSEND ENDS
_STACK  SEGMENT STACK 'STACK'
_STACK  ENDS
_INIT   SEGMENT BYTE PUBLIC 'STACK'  ;used for initialization code
_INIT   ENDS

DGROUP  GROUP   _TEXT, _DATA, _BSS, _BSSEND, _STACK, _INIT

        ASSUME  CS:DGROUP
;=======================================================================
MSDOS   MACRO   name            ;MSDOS INT 21h calls
        IFDEF   &name
          MOV   AH,&name
        ENDIF
        INT     DOS
ENDM
;---------------------------------------------------------------------------
; most of the TSR program functions must be performed during interrupt
; processing, and there may not be enough stack space available. Hence we
; swap to a local stack, but protect it against multiple use! Note that
; this assumes a single CPU.

stack_switch macro old_ss, old_sp, busy, stack, skip
        cmp     B DG:busy,1
        je      skip
        cli
        mov     B DG:busy,1
        mov     W DG:old_sp,sp
        mov     W DG:old_ss,ss
        mov     ax,cs
        mov     ss,ax
        mov     sp,OS DG:stack
        sti
endm
;---------------------------------------------------------------------------
stack_restore macro old_ss, old_sp, busy
        cli
        mov     ss,W DG:old_ss
        mov     sp,W DG:old_sp
        mov     B DG:busy, 0
endm
;---------------------------------------------------------------------------
out_al  macro   address, value
        mov     dx,W DG:address
        mov     al,value
        out     dx,al
endm
;---------------------------------------------------------------------------
in_al   macro   address
        mov     dx,W DG:address
        in      al,dx
endm
;=======================================================================
_TEXT   SEGMENT

;Our service routine for the INT 2F chain to check, if TSR is already loaded.
;At entry: AH signals, if our program is asked for his signature
;if yes, we deliver the pointer ES:BX to our signature

my2F    proc    far
        CMP     AH,B DG:id_2FH  ;does AH meet our ID number?
        JZ      us              ;yes
        JMP     D CS:old_2F     ;no, chain to the old vector
us:     PUSH    CS
        POP     ES              ;ES:BX point to segment and
        MOV     BX,OS m1        ;  offset of our logo (signature)
        MOV     AL,0FFH         ;flag: success
        IRET
my2F    endp

;ATTENTION: ES is 256 bytes above the PSP (important for FREEMEM)
;---------------------------------------------------------------------------
if VIDEO_ISR
; this is the video interrupt (probably BIOS VIDEO call)
video_int       proc    far
        pushf
        call    D cs:old_video
                                ;tests to be put here
        iret
video_int       endp
endif
;---------------------------------------------------------------------------
kb_handler      proc    far     ;hardware KBD INT 9 service routine
        push    ax
        push    ds
        mov     ax,cs
        mov     ds,ax           ;now we can access our flags

      ; mov     al, INT_ENABLE  ;no service for interrupt controller
      ; out     INT_CTL, al     ; necessary! ?

        pushf
        call    D cs:old_keyboard ;first let the BIOS read the KBD

        mov     ax,1100h
        int     KBD_BIOS        ;char in KBD Buffer available?
        jz      kb_exit         ;no, do nothing
        cmp     ah,B DG:_g_hot_scan ;yes, scan code of trigger key?
        jne     kb_exit         ;no
        mov     ax,1200h
        int     KBD_BIOS
        and     al,B DG:_g_hot_ctrl ;correct additional key pressed?
        je      kb_exit         ;no, leave keystroke in buffer
        mov     ax,1000h        ;yes, it's our trigger key
        int     KBD_BIOS        ;take keystroke out of the buffer
        not     B DG:pause_scan ;negate the pause flag
        call    disp_mode       ;get actual display mode

kb_exit:pop     ds
        pop     ax
        iret
kb_handler      endp
;---------------------------------------------------------------------------
; this is the hardware timer interrupt. The total screen is checked once
mytimer proc    far

        push    ax
        push    ds
        push    bx
        push    dx
        mov     ax, cs
        mov     ds, ax          ;now we have our ds

   if PULSES                    ; begin pulse A: before MSDOS service
        mov     bl,01h          ; set puls on bit 0
        pushf
        pop     ax
        test    ax,INTR_EN_BIT  ; INTRs enabled?
        je      lo_p_myti1      ; no, low puls
        or      bl,02h          ; yes, high puls (bit 1 added)
lo_p_myti1:
        in_al   oscope_port ; read present bits
        or      al,bl           ; and add new bits
        out     dx,al
   endif

        pushf
        call    D cs:old_timer  ;first service other routines

   if PULSES                    ; begin pulse B: after end of MSDOS service
        mov     bl,04h          ; set puls on bit 2
        pushf
        pop     ax
        test    ax,INTR_EN_BIT  ; INTRs enabled?
        je      lo_p_myti2      ; no, low puls
        or      bl,08h          ; yes, high puls (bit 3 added)
lo_p_myti2:
        in_al   oscope_port
        or      al,bl
        out     dx,al
   endif

        cmp     B DG:pause_scan,0 ;screen scanning allowed?
        je      do_nothing      ;no
        cmp     B DG:ticks, 0   ;wait count exhausted?
        jg      do_nothing      ;no
        mov     al, B DG:_g_delay_cnt
        mov     B DG:ticks,al   ;yes, reload counter

        stack_switch big_ss, big_sp, big_busy, big_stack, do_nothing

        push    cx              ;this code is performed only, if
        push    si              ;previous interrupt has finished
        push    di
        push    bp
        push    es

   if PULSES                    ; update pulse B
        pushf
        pop     ax
        test    ax,INTR_EN_BIT  ; INTRs enabled?
        je      lo_p_myti3      ; no, low puls
        in_al   oscope_port
        or      al,08h          ; set puls on bit 3
        out     dx,al
lo_p_myti3:
   endif

        call    _work_on_screen ; the main work is done by the C module!

   if PULSES                    ; update pulse B (2)
        pushf
        pop     ax
        test    ax,INTR_EN_BIT  ; INTRs enabled?
        je      lo_p_myti4      ; no, low puls
        in_al   oscope_port
        or      al,08h          ; set puls on bit 3
        out     dx,al
lo_p_myti4:
   endif

        pop     es
        pop     bp
        pop     di
        pop     si
        pop     cx
        stack_restore big_ss, big_sp, big_busy

do_nothing:
   if PULSES                    ; update pulse B (3)
        mov     bl,04h          ; assume INTR bit disabled
        pushf
        pop     ax
        test    ax,INTR_EN_BIT  ; INTRs enabled?
        je      lo_p_myti5      ; no, low puls
        or      bl,08h          ; yes, set bit 3
lo_p_myti5:
        in_al   oscope_port
        and     al,0f3h         ; reset bit 3 + 4
        or      al,bl           ; set actual status
        out     dx,al
   endif

        dec     B DG:ticks

   if PULSES                    ; clear pulses A and B
        mov     bl,0f0h         ; reset bit 0 t0 3
        in_al   oscope_port
        and     al,bl
        out     dx,al
   endif

        pop     dx
        pop     bx
        pop     ds
        pop     ax

        iret
mytimer endp
;---------------------------------------------------------------------------
;someone could have switched by the DOS commands: MODE MONO / MODE CO80 !!
;  (e.g. for performing dual screen debugging with the turbo debugger)
disp_mode       proc    near
        mov     ah, DISPLAY_MODE ; Find out if color or mono screen
        int     VIDEO
        cmp     al, 7
        jz      mono_screen
        mov     W DG:screen_seg, COLOR_BASE
        jmp     short disp_ex
mono_screen:
        mov     W DG:screen_seg, MONO_BASE
disp_ex:
        ret
disp_mode       endp
;---------------------------------------------------------------------------
_check_lines    proc    near
        cmp     W DG:screen_seg, COLOR_BASE ;actual screen mode
        jnz     chk1
        xor     ax,ax           ;in color mode we ask the BIOS data area
        mov     es,ax
        mov     bx,0484h        ;es:bx --> address 40h:84h
        mov     al,es:bx
        mov     W DG:_g_num_lines,ax
        ret
chk1:   mov     W DG:_g_num_lines,24 ; mono base has always 25 lines
        ret
_check_lines    endp

;---------------------------------------------------------------------------
_hilite_string  proc near       ;reverse marking
        push    di              ;SI/DI are REGISTER variables in C !!
        mov     ax,W DG:screen_seg
        mov     es,ax
        mov     di,W DG:_g_temp_cell ;lower cell
        inc     di
        mov     al,B DG:_g_newattr
        mov     bx,W DG:_g_current_cell ;higher cell
        inc     bx
        cld

        cli                     ;must be, otherwise IDE of QBasic blocks
hloop:  stosb
        inc     di
        cmp     bx,di
        ja      hloop
        sti

        pop     di
        ret
_hilite_string  endp
;---------------------------------------------------------------------------
_remark_string  proc near       ;forward marking
        push    di
        mov     ax,W DG:screen_seg
        mov     es,ax
        mov     di,W DG:_g_current_cell
        inc     di
        mov     al,B DG:_g_newattr
        mov     bx,W DG:_g_temp_cell
        inc     bx
        cld

        cli                     ;must be, otherwise IDE of QBasic blocks
rloop:  stosb
        inc     di
        cmp     bx,di
        ja      rloop
        sti

        pop     di
        ret
_remark_string  endp
;---------------------------------------------------------------------------
_array_scan     proc    near

        push    ds              ;ax,bx,cx,dx,es are free in C
        push    si              ;SI/DI are REGISTER variables in C !!
        push    di

        mov     ax,cs
        mov     es,ax
        mov     di,OS DG:_g_kwrd_buf ;es:di = prototype array pointer (dest)
        mov     dx,di                ;dx: current offset in prototype array
        Mov     si,OS DG:_g_temp_str ;ds:si = temp array pointer (source)
        cld

        mov     ax,W DG:_g_maxwords
        push    bp
        mov     bp,sp
        push    ax              ;save MaxWord% at bp-2

        xor     ax,ax
        mov     al,byte ptr ds:[si]
        mov     bx,ax           ;bx: length of temp string to be searched
        inc     si
        push    si              ;addr of 1st char in temp string

cmplen: xor     ax,ax
        mov     al,byte ptr es:[di] ;ax: length of actual prototype string
        cmp     ax,bx           ;temp and actual, equal lengths?
        jne     nextproto       ;no

        pop     si              ;yes, look if strings are equal:
        push    si              ;back to start of temp string
        inc     di
        inc     di              ;start of actual prototype
        mov     cx,bx           ;cx = repeat count
        repe cmpsb              ;strings matching?
        jne     nextproto       ;no

        mov     di,dx           ;back to start of actual prototype
        inc     di              ;position of group identifier
        xor     ax,ax
        mov     al,byte ptr es:[di] ;ax: group of string found
        jmp     short cmpfini2

nextproto:
        dec     word ptr [bp-2] ;MaxKeyWords exhausted?
        je      cmpfini1        ;yes, nothing found
        add     ax,2            ;add extra bytes to axtual length
        add     dx,ax           ;new start offset in Prototype array
        mov     di,dx
        jmp     short cmplen

cmpfini1:
        xor     ax,ax           ;ax=0 --> nothing found
cmpfini2:
        add     sp,4
        pop     bp
        pop     di
        pop     si
        pop     ds

        ret                     ;ax contains the group number
_array_scan     endp
;---------------------------------------------------------------------------

_TEXT   ENDS
;=======================================================================
_DATA   SEGMENT

pause_scan      db      0ffh    ;pause flag for keyword scan
id_2FH  DB      0               ;code of our ID number
m1      DB      18 dup (32)
        db      '(ASM/C  version) - (c) Andras Hoeffken '
        db      '<ah@confusion.rmc.de>',CR,LF,'$' ;our logo and INT 2F ID
end_m1  label byte
logolen EQU   end_m1-m1         ;taken as signature for the INT 2Fh test!

old_2F          dw      0,0     ; original int 2F handler (ofs,segm)
if VIDEO_ISR
  old_video     dw      0,0     ; original VIDEO interrupt handler
endif
old_keyboard    dw      0,0     ; original KEYBOARD interrupt handler
old_timer       dw      0,0     ; original timer interrupt handler
big_ss          dw      0       ; for saving old SS and SP values
big_sp          dw      0
big_busy        db      1 ; stack busy flag (1 = stop during TSR installation)

_g_phys_screen  label   dword   ;address of top left corner of text screen
screen_off      dw      0       ; (an initialized pointer variable)
screen_seg      dw      MONO_BASE ;start of video RAM (text mode)
_g_num_lines    dw      0       ;actual number of screen lines

_g_maxwords     dw      0       ;total number of keywords in array
_g_bcount       dw      0       ;length of keywordbuffer in bytes
_g_k_colors     label byte      ;array for 10 keyword colors
                db      10 dup (0)
_g_temp_str     label byte      ;temporary string for keyword scan
                db      80 dup (0) ; max 1 screen line
_g_newattr      db      0
_g_temp_cell    dw      0
_g_current_cell dw      0


if PULSES
oscope_port     dw      0278h   ; LPT2 port to connect an oscilloscope
endif

ticks           db      0ffh    ; timer ticks to wait before hilite action

DATA   ENDS
;==========================================
_BSS    SEGMENT
bdata@  label   byte
_BSS    ENDS
;==========================================
_BSSEND SEGMENT
edata@  label   byte
_BSSEND ENDS
;==========================================
_STACK  SEGMENT
        dw      128 dup (?)
big_stack       label byte
stack_end       label byte
_STACK  ENDS
;=======================================================================
_INIT   SEGMENT

_g_hot_scan     db      1       ; to switch mytimer service on/off
_g_hot_ctrl     db      4       ; 4 = CTRL pressed
_g_delay_cnt    db      2       ; hilite actions after 2 timer ticks
                db      0       ; (for even address alignement)

_g_kwrd_buf     label   byte
                db      KEYWORD_MEMORY dup (0) ;keyword buffer

def_file_name   db      'PQSYNTAX.DEF',0,':$'
                db      30 dup (0) ;for long pathnames
f_handle        dw      0       ;file handle
fpos_low        dw      0       ;position of file pointer
fpos_high       dw      0
_g_f_len        dw      0       ;length of file
_g_test         db      0       ;1 = do only test the DEF file
                                ;0 = normal TSR performance
;---------------------------------------------------------------------------
m0  DB CR,LF,'PQSYNTAX Vs 1.4 - Highlighting keywords in the IDE of'
    db ' PowerBASIC/QuickBasic/PDS',CR,LF,'$'
m2  DB 'Can not install PQSYNTAX !!'
    DB ' - Program is already loaded',CR,LF,'$'
m3  DB 'The program has been loaded resident. '
    DB 'To unload, type "PQSYNTAX /U"',CR,LF,'$'
m4  DB 'PQSYNTAX is not present',CR,LF,'$'
m6  DB 'Can not unload PQSYNTAX',CR,LF,'$'
m7  DB 'PQSYNTAX is being unloaded',CR,LF,'$'
m8  DB 'Cannot free ENV space, exit',CR,LF,'$'
m9  DB 'Error on command line, exit',CR,LF,'$'
m10 DB 'Cannot open DEF file',CR,LF,'$'
m11 DB 'Cannot get length of DEF file',CR,LF,'$'
m12 DB 'DEF file is too long (max 8 KB!)',CR,LF,'$'
m13 DB 'Error reading DEF file'
m14 db CR,LF,'$'
m15 DB 'Usage: PQSYNTAX <options>',CR,LF
db '                nothing    - run once (testing default PQSYNTAX.DEF file)',CR,LF
db '                /F name    - run once (testing user DEF file)',CR,LF
db '                /R         - go resident using default PQSYNTAX.DEF file',CR,LF
db '                /R /F name - go resident using user DEF file',CR,LF
db '                /U         - unload the TSR program',CR,LF
    db 'Evaluation of DEF file $'
_g_z0 label byte
db 'Trigger Key: Scancode xx / Shiftstatus xx - Delaycount xx',CR,LF,CR,LF,'$'
;---------------------------------------------------------------------------
; this code is inserted here so that it does not waste any memory
; startup code - this is where execution begins

startup         PROC    NEAR
        mov     dx, cs          ; first: adjust the segment
        mov     ds, dx          ; registers DS,SS,ES to CS
        mov     es, dx
        mov     ax, OS DG: big_stack
        cli
        mov     ss, dx
        mov     sp, ax
        sti

        xor     ax, ax          ; zero uninitialized data area (BSS segm)
        mov     es, dx          ; (not used here, but useful for general)
        mov     di, OS DG: bdata@
        mov     cx, OS DG: edata@
        sub     cx, di
        rep     stosb

        call    main            ;(sets al = exit code)
        MSDOS   TERMINATE
startup         ENDP
;---------------------------------------------------------------------------
main    proc    near
        MOV     DX,OS DG:m0     ;title
        MSDOS   DISPLAY_STRING  ;Display string at (DS:DX)
        MOV     DX,OS DG:m1     ;and ID
        MSDOS   DISPLAY_STRING  ;Display string at (DS:DX)
        MOV     AX,CS           ;free unused ENV space
        SUB     AX,10H
        MOV     ES,AX           ;ES points to PSP
        MOV     ES,W ES:[ENV]   ;fetch pointer to ENV memory
        MSDOS   FREEMEM         ;get rid of allocated ENV memory
        JNC     main_1          ;jmp if ok
        MOV     DX,OS DG:m8     ;no (cannot free ENV ..)
        MSDOS   DISPLAY_STRING  ;Display string at (DS:DX)
        RET                     ;otherwise exit to DOS

cmd_line_error:
        MOV     DX,OS DG:m9     ;no, (no uninstall possible)
        MSDOS   DISPLAY_STRING  ;Display string at (DS:DX)
        MOV     AL,CMD_ERROR    ;mark error
        RET

main_1: MOV     AX,CS
        SUB     AX,10H
        MOV     ES,AX           ;ES points to PSP
        MOV     SI,LEN_P_ST     ;point to DOS command buffer
        xor     cx,cx
        MOV     cl,B ES:[SI]    ;fetch length of cmd string
        or      cl,cl
        jnz     main_2
        inc     B DG:_g_test    ;no command means mode: test the DEF file
        jmp     install         ; (here: use default DEF file)

main_2: inc     cx              ;correct length
        call    get_char
        jc      cmd_line_error
        cmp     al,'/'          ;first char must be a /
        jne     cmd_line_error

        call    get_char
        jc      cmd_line_error
        cmp     al,'U'          ;2nd Char must be U=uninstall
        je      uninst
        cmp     al,'R'          ;  or R=resident
        je      main_3
        cmp     al,'F'          ;  or F=File
        jne     cmd_line_error
        inc     B DG:_g_test    ;no command R/U means mode: test the DEF file
        jmp     short main_5    ; (here: use user DEF file)

main_3: call    get_char
        jnc     main_4
        jmp     install         ;no more chars: default is PQSYNTAX.DEF
main_4: cmp     al,'/'          ;we wait for another /
        jne     cmd_line_error
        call    get_char
        jc      cmd_line_error
        cmp     al,'F'          ;then an F=file must follow
        jne     cmd_line_error
main_5: call    get_char        ;position to 1st byte of "path\filename"
        cmp     cx, 40          ;max 40 chars for user "path\filename"
        ja      cmd_line_error

        push    ds
        mov     ax,es
        push    ds
        pop     es
        mov     ds,ax                  ;ds:si = source
        mov     di,OS DG:def_file_name ;es:di = destination
        rep     movsb           ;copy the user filename
        xor     al,al
        stosb                   ;terminate string with 0
        mov     ax,'$:'
        stosw                   ;add 2 bytes to prepare screen output
        pop     ds
        jmp     short install   ;install the TSR program

uninst: CALL    ptest           ;UNINSTALL: test for presence
        OR      AH,AH           ;present (ah <> 0)?
        JNZ     unin_1          ;yes, uninstall seems possible
        MOV     DX,OS DG:m4     ;no, (no uninstall possible)
        MSDOS   DISPLAY_STRING  ;Display string at (DS:DX)
        MOV     AL,TSR_ERROR    ;mark error
        RET

;uninstall is only done, if the TSR, which was found, is the last
;loaded program of the INT 2Fh chain!!

unin_1: PUSH    ES              ;segment of prog found by 'ptest'

        MOV     AH,GETVECT      ;Get actual INT 2Fh Vector ES:BX
        MOV     AL,MUPLEX       ;  (at the end of the chain!!)
        MSDOS

        POP     AX              ;AX = former ES
        MOV     CX,ES           ;CX = segment of INTR-routine
        CMP     AX,CX           ;do we own (this last) TSR-prog?
        JNZ     unin_3          ;no
        MOV     AX,OS DG:my2F   ;AX = our offset
        CMP     AX,BX           ;same as BX = offs of INTR-rout?
        JNZ     unin_3          ;no
        MOV     DX,OS DG:m7     ;yes (TSR is beeing unloaded)
        MSDOS   DISPLAY_STRING  ;Display string at (DS:DX)

        MOV     AH,SETVECT      ;restore old INT 2Fh Vector
        MOV     AL,MUPLEX
        mov     bx,OS DG:old_2F
        MOV     DX,W ES:bx      ;original offset
        mov     bx,OS DG:old_2F+2
        MOV     DS,W ES:bx      ;original segment
        MSDOS

        call    deinst_other_vectors

        MOV     AX,ES
        SUB     AX,10H          ;important for our .EXE program!
        MOV     ES,AX           ;ES points to start of prog (PSP)
        MSDOS   FREEMEM         ;Free our allocated memory
        JC      unin_2          ;carry = error (al = error code)
        MOV     AL,0            ;mark: al = 0
unin_2: RET

unin_3: MOV     DX,OS DG:m6     ;(Cannot unload ..)
        MSDOS   DISPLAY_STRING  ;Display string at (DS:DX)
        MOV     AL,TSR_ERROR
        RET

install:cmp     B DG:_g_test,1  ;test mode?
        jne     inst_0          ;no
        MOV     DX,OS DG:m15    ;yes, display test info
        MSDOS   DISPLAY_STRING
        MOV     DX,OS DG:def_file_name
        MSDOS   DISPLAY_STRING
        MOV     DX,OS DG:m14
        MSDOS   DISPLAY_STRING
        MOV     DX,OS DG:m14
        MSDOS   DISPLAY_STRING
        jmp     short inst_2
inst_0: CALL    ptest           ;test if our prog is present
        OR      AH,AH           ;already present (AH <>0)?
        JZ      inst_1          ;no, we can install our prog
        MOV     DX,OS DG:m2     ;yes (Cannot install ..)
        MSDOS   DISPLAY_STRING  ;Display string at (DS:DX)
        MOV     AL,TSR_ERROR    ;mark error
        RET

inst_1: MOV     B DG:id_2FH,AL  ;keep our ID code in mind

inst_2: call    store_file      ;read file with definition of keywords
        jnc     inst_3
        mov     al,1            ;error reading file, exit
        ret
inst_3: call    _eval_file      ;evaluate DEF file

   if TSR
        cmp     B DG:_g_test,1  ;proceed with installation?
        jne     inst_4          ;yes
        xor     ax,ax           ;no, exit DEF file check with error flag = 0
        ret
   endif

inst_4: call    disp_mode       ;check actual displaymode

debuglabel:
   if TSR eq 0                  ;for testing (no TSR):
        call    _check_lines
        call    _hilite_screen  ;run once through a 25x80 screen
        mov     al,0            ;error 0
        ret                     ;finish, exit to DOS
   endif

        mov     ah, GETVECT     ; save original MULTIPLEX interrupt
        mov     al, MUPLEX
        int     DOS
        mov     W DG:old_2F,bx  ; offset
        mov     W DG:old_2F+2,es; segment

        call    inst_other_vectors;install rest of our INT service routines

        mov     dx, OS DG:my2F  ; install replacement
        mov     ah, SETVECT     ; MULTIPLEX interrupt handler
        mov     al, MUPLEX
        int     DOS

        MOV     DX,OS DG:m3     ;(TSR installed)
        MSDOS   DISPLAY_STRING  ;Display string at (DS:DX)
        MOV     AL,0            ;error code = 0
        MOV     AH,MAKE_TSR     ;Terminate and Stay Resident

                                     ;up to here we keep our program:
        mov     dx,OS DG:_g_kwrd_buf ; start of keyword buffer plus
        add     dx,W DG:_g_bcount    ; length of keyword buffer (shortest TSR)

        ADD     DX,0FH          ;add 0F for segment boundary
        ADD     DX,100H         ;add 100H for PSP
        SHR     DX,1            ;/2
        SHR     DX,1            ;/4
        SHR     DX,1            ;/8
        SHR     DX,1            ;/16 = paragraphs to keep
        mov     byte ptr DGROUP:big_busy, 0  ;allow mytimer routine to work
        MSDOS                   ;TSR installation finished

main    endp

;---------------------------------------------------------------------------
get_char proc near              ;read one char from cmd line
        dec     cx
        je      get_e           ;cmd line exhausted
        inc     si
        mov     al,ES:[si]
        cmp     al,' '          ;move over spaces
        je      get_char
        cmp     al,97           ; a
        jl      get_ok
        cmp     al,122          ; z
        jg      get_ok
        sub     al,32           ;make upper case letters
get_ok: clc
        ret
get_e:  stc
        ret
get_char endp
;-----------------------------------------------------------------------------
;test if a program is present within the chain of INT 2Fh
;multiplex interrupt. Conventions:
;AH = multiplex ID code number, individual for a single program
;  (AH=00..BF reserved for DOS! AH=C0..FF for user programs!!)
;AL = flag, delivered after an INT 2Fh:
;  AL=FFh --> the program with the specific AH-code has been found
;  AL=00  --> no success
;in the following 'ptest' ALL user IDs (C0..FF) are tested!
;if our program seems to be found, a string compare is made for
;security.

ptest   proc    near
        PUSH    DI              ;look (by INT 2Fh), if DOS has a
        PUSH    SI              ; notation of our TSR-program
        PUSH    BX
        PUSH    DX
        MOV     DH,0BFH         ;user code numbers: C0h...FFh
        MOV     DL,0            ;reset for later
ptes_1: ADD     DH,1            ;try the next code (C0..FF..00)
        JZ      ptes_3          ;00: jmp, if range exhausted
        MOV     AH,DH           ;the actual code to be tested
        MOV     AL,0            ;reset al (= result flag)
        MOV     BX,0            ;reset BX (offset of dest.)
        MOV     ES,BX           ;reset ES (segment of dest.)
        PUSH    DX
        INT     2FH             ;look if a notation for the
        POP     DX              ;  actual code is found
        CMP     AL,0FFH         ;code in use?
        JZ      ptes_2          ;(al = 0FFh ->) seems to be
        OR      DL,DL           ;no, DL still 0?
        JNZ     ptes_1          ;no, continue
        MOV     DL,DH           ;yes, store 1st empty 2FH_ID code
        JMP     SHORT ptes_1    ;continue looking for our prog
ptes_2: MOV     AX,ES           ;ES = segment pointer:
        OR      AX,AX           ;  value reported?
        JZ      ptes_1          ;no, continue searching
        MOV     DI,BX           ;yes, ES:DI --> the other logo
        MOV     SI,OFFSET m1    ;Source = our logo (DS:SI)
        MOV     CX,OFFSET logolen ;length of the logo
        REPZ CMPSB              ;Compare strings
        JNZ     ptes_1          ;continue searching, if not ours
ptes_3: MOV     AX,DX           ;AH flags result (0 if not found)
        POP     DX              ;AL gives the first free ID code
        POP     BX
        POP     SI              ;ES has segment of found TSR
        POP     DI
        RET                     ;present test finished
ptest   endp
;---------------------------------------------------------------------------
inst_other_vectors proc near
    if VIDEO_ISR
        mov     ah, GETVECT     ; save original VIDEO interrupt
        mov     al, VTRIGGER
        int     DOS
        mov     W DG:old_video,bx
        mov     W DG:old_video+2,es
    endif
        mov     ah, GETVECT     ; save original KEYBOARD interrupt
        mov     al, KBD_HARD
        int     DOS
        mov     W DG:old_keyboard,bx
        mov     W DG:old_keyboard+2,es

        mov     ah, GETVECT     ; save original TIMER interrupt
        mov     al, TIMER
        int     DOS
        mov     W DG:old_timer,bx
        mov     W DG:old_timer+2,es

         ;(ds is already set correctly!)
        mov     dx, OS DG:mytimer ; install replacement
        mov     ah, SETVECT       ; TIMER interrupt handler
        mov     al, TIMER
        int     DOS

        mov     dx, OS DG:kb_handler ; install replacement KEYBOARD interrupt
        mov     ah, SETVECT          ; handler - note that the keyboard
        mov     al, KBD_HARD    ; interruptcannot be masked off or
        int     DOS             ; there are problems with CAPS LOCK etc
    if VIDEO_ISR
        mov     dx, OS DG:video_int ; install replacement
        mov     ah, SETVECT         ; VTRIGGER interrupt handler
        mov     al, VTRIGGER
        int     DOS
    endif
                ret
inst_other_vectors endp
;---------------------------------------------------------------------------
deinst_other_vectors proc near
    if VIDEO_ISR
        MOV     AH,SETVECT      ;restore old Video INT Vector
        MOV     AL,VTRIGGER
        mov     bx,OS DG:old_video
        MOV     DX,W ES:bx      ;original offset
        mov     bx,OS DG:old_video+2
        MOV     DS,W ES:bx      ;original segment
        MSDOS
    endif
        MOV     AH,SETVECT      ;restore old keyboard INT Vector
        MOV     AL,KBD_HARD
        mov     bx,OS DG:old_keyboard
        MOV     DX,W ES:bx      ;original offset
        mov     bx,OS DG:old_keyboard+2
        MOV     DS,W ES:bx      ;original segment
        MSDOS

        MOV     AH,SETVECT      ;restore old timer INT Vector
        MOV     AL,TIMER
        mov     bx,OS DG:old_timer
        MOV     DX,W ES:bx      ;original offset
        mov     bx,OS DG:old_timer+2
        MOV     DS,W ES:bx      ;original segment
        MSDOS

        ret
deinst_other_vectors endp
;---------------------------------------------------------------------------
store_file      proc    near
        mov     ax, 3D00h       ;OPEN FILE (AL = 0: only FOR READ)
        mov     dx, OS DG:def_file_name ;DS:DX = STRING pointer
        INT     DOS
        jnc     LB2             ;carry flag set IF ERROR
        MOV     DX,OS DG:m10    ;(cannot open ...)
        MSDOS   DISPLAY_STRING  ;Display string at (DS:DX)
        stc                     ;carry = error
        ret

LB2:    mov     W DG:f_handle, ax ;file handle
        mov     ax, 4200h + 1   ; get current file pointer position
        mov     bx, W DG:f_handle
        mov     cx, 0           ; hi-WORD of offset
        mov     dx, 0           ; lo-WORD of offset
        INT     DOS
        jc      LB7             ;error
        mov     W DG:fpos_low, ax  ; preserves current file pointer
        mov     W DG:fpos_high, dx ;   position

        mov     ax, 4200h + 2   ; get file size (pointer to EOF)
        mov     bx, W DG:f_handle
        mov     cx, 0           ; hi-WORD of offset
        mov     dx, 0           ; lo-WORD of offset
        INT     DOS
        jc      LB7             ;error

        push    dx              ; hi-WORD of filelength
        push    ax              ; lo-WORD
        mov     dx, W DG:fpos_low  ; RESTORE previous file pointer
        mov     cx, W DG:fpos_high ;   position
        mov     ax, 4200h
        mov     bx, W DG:f_handle
        INT     DOS
        pop     ax
        pop     dx
        jc      LB7             ;error

        or      dx,dx           ;check file length:
        jne     LB6
        cmp     ax,KEYWORD_MEMORY ;too long?
        jg      LB6             ;yes

        mov     cx,ax           ;no, read cx no of bytes
        mov     W DG:_g_f_len,ax ;keep length in mind
        mov     ax, 3F00h
        mov     bx, W DG:f_handle
        mov     dx, OS DG:_g_kwrd_buf ;DS:DX = buffer pointer
        INT     DOS
        jc      LB5             ; carry flag if read ERROR
        mov     ax, 3E00h       ;CLOSE FILE
        mov     bx, W DG:f_handle
        INT     DOS
        clc                     ;ok flag
        ret

LB5:    MOV     DX,OS DG:m13    ;(cannot read ..)
        MSDOS   DISPLAY_STRING  ;Display string at (DS:DX)
        jmp     short LB8

LB6:    MOV     DX,OS DG:m12    ;(file too long ..)
        MSDOS   DISPLAY_STRING  ;Display string at (DS:DX)
        jmp     short LB8

LB7:    MOV     DX,OS DG:m11    ;(cannot move ...)
        MSDOS   DISPLAY_STRING  ;Display string at (DS:DX)
LB8:    mov     ax, 3E00h       ;CLOSE FILE
        mov     bx, W DG:f_handle
        INT     DOS

        stc                     ;carry = error
        ret
store_file      endp
;---------------------------------------------------------------------------

_INIT   ENDS
;==========================================
;PUBLIC declarations

;for PQHILITE.C                     ;functions:
        public  _check_lines
        public  _hilite_string
        public  _remark_string
        public  _array_scan
                                    ;variables:
        public  _g_phys_screen
        public  _g_num_lines
        public  _g_temp_str
        public  _g_newattr
        public  _g_temp_cell
        public  _g_current_cell

;for PQREADFL.C                     ;variables:
        public  _g_maxwords
        public  _g_bcount
        public  _g_hot_scan
        public  _g_hot_ctrl
        public  _g_delay_cnt
        public  _g_kwrd_buf
        public  _g_f_len
        public  _g_test
        public  _g_z0

;for both PQHILITE.C + PQREADFL.C   ;variable:
        public  _g_k_colors

;for test purpose (.MAP file)
   if TSR eq 0
        public  my2F
        public  kb_handler
        public  mytimer
        public  old_2F
        public  ticks
        public  stack_end
        public  startup
        public  unin_1
        public  debuglabel
        public  disp_mode
        public  store_file
        public  main_1
        public  def_file_name
   endif

        END     startup

